import ast
from pathlib import Path

from django_translate_gettext.constants import TO_SKIP
from django_translate_gettext.services.transformers import ClassDefTransformer


def fetch_app_files(app_name: str) -> set[Path]:
    """Fetch all python files in the app directory excluding the files in the TO_SKIP list.

    Args:
        app_name (str): The app name to fetch the files from.

    Returns:
        set[Path]: set of filtered Pathlib objects for the files in the app.
    """
    return {
        file
        for file in Path(app_name).rglob("*.py")
        if file.is_file() and not any(skip in str(file) for skip in TO_SKIP)
    }


def update_py_file(*, file_path: Path) -> None:
    """Update the python file with the gettext call wrapping.

    Args:
        file_path (Path): The file path Pathlib object to update.

    Returns:
        None
    """
    tree = ast.parse(file_path.read_text())

    transformer = ClassDefTransformer()
    new_tree = transformer.visit(tree)
    new_tree = transformer.insert_getetxt_import(new_tree)

    code = ast.unparse(new_tree)
    file_path.write_text(code)



from django.db.models import Model


def get_all_app_models(app_label: str) -> set[Model]:
    """Get all models for the app label including the abstract models.

    Args:
        app_label (str): The app label to get the models for.

    Returns:
        set[Model]: The set of models for the app label.
    """
    result = set()
    generation = {Model}
    while generation:
        generation = {sc for c in generation for sc in c.__subclasses__()}
        result.update([c for c in generation if c._meta.app_label == app_label])  # noqa: SLF001
    return result



import concurrent.futures
import subprocess
from contextlib import suppress
from pathlib import Path
from typing import NamedTuple

from django.apps import apps
from django.core.management.base import BaseCommand

from django_translate_gettext.exceptions import TranslatorError
from django_translate_gettext.services import update_py_file
from django_translate_gettext.services.files import fetch_app_files
from django_translate_gettext.services.translators import PoFileTranslator

MAX_WORKERS = 5


class FileToGettext(NamedTuple):
    file_path: Path
    formatted: bool


class Command(BaseCommand):
    help = "Add calling gettext for the model files"

    def add_arguments(self, parser) -> None:
        parser.add_argument(
            "apps", nargs="+", type=str, help="Apps to add gettext for the model files.\nFor example: app1 app2 app3"
        )
        parser.add_argument(
            "-f",
            "--format",
            action="store_true",
            help="Call Ruff formatting tool to format the code after generating new model files.",
        )
        parser.add_argument(
            "-mm",
            "--makemessages",
            nargs="+",
            type=str,
            help="Call makemessages command to create the .po files after adding the gettext "
            "and translating for the passed languages."
            "\nFor example: en de fr",
        )

    def handle(self, **options) -> None:
        files_to_gettext = []
        for app_name in options["apps"]:
            try:
                apps.get_app_config(app_name)
            except LookupError as error:
                self.stdout.write(self.style.ERROR(error))
                continue

            files_to_gettext.extend(self.fetch_app_files_to_gettext(app_name=app_name, formatted=options["format"]))

        self.stdout.write(
            self.style.WARNING(
                "Please, check the files, and don't forget to call migrate command to apply the changes "
                "to the database after adding the gettext."
            )
        )

        self.add_gettext_for_files(files=files_to_gettext)
        self.stdout.write(self.style.SUCCESS("Successfully added gettext for the apps files."))

        self.process_translating(**options)

    def translate_lang_code(self, lang_code: str) -> None:
        translator = PoFileTranslator(lang_code=lang_code)
        translator.translate_codes()
        self.stdout.write(self.style.SUCCESS(f"Successfully translated for lang code {lang_code}."))

    def process_translating(self, **options) -> None:
        if not options["makemessages"]:
            self.stdout.write(
                self.style.SUCCESS(
                    "Please, un the command 'python manage.py makemessages -l {lang code}' to create the .po files."
                )
            )
            return

        self.stdout.write(self.style.WARNING("Calling makemessages command to create the .po files."))
        langs = [f"--locale={lang}" for lang in options["makemessages"]]
        with suppress(subprocess.CalledProcessError):
            subprocess.run(["python", "manage.py", "makemessages", *langs], check=True)  # noqa: S603, S607

        lang_codes = options["makemessages"]
        max_workers = MAX_WORKERS if len(lang_codes) > MAX_WORKERS else len(lang_codes)
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [executor.submit(self.translate_lang_code, code) for code in lang_codes]
            for future in concurrent.futures.as_completed(futures):
                try:
                    future.result()
                except TranslatorError as error:  # noqa: PERF203
                    self.stdout.write(self.style.ERROR(f"Translator error: {error}"))

    @staticmethod
    def fetch_app_files_to_gettext(*, app_name: str, formatted: bool = False) -> list[FileToGettext]:
        return [
            FileToGettext(file_path=file_path, formatted=formatted) for file_path in fetch_app_files(app_name=app_name)
        ]

    @staticmethod
    def gettext_py_file(file_path: Path, formatted: bool = False) -> None:  # noqa: FBT001, FBT002
        with suppress(FileNotFoundError):
            update_py_file(file_path=file_path)
            filename = file_path.absolute()
            if formatted:
                subprocess.run(["ruff", "format", f"{filename!s}"], check=True)  # noqa: S603, S607

    def add_gettext_for_files(self, files: list[FileToGettext]) -> None:
        max_workers = MAX_WORKERS if len(files) > MAX_WORKERS else len(files)
        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
            futures = [executor.submit(self.gettext_py_file, f.file_path, f.formatted) for f in files]
            for future in concurrent.futures.as_completed(futures):
                try:
                    future.result()
                except FileNotFoundError as error:  # noqa: PERF203
                    self.stdout.write(self.style.ERROR(f"File not found error: {error}"))



import ast
from _ast import stmt
from contextlib import suppress


class ClassDefTransformer(ast.NodeTransformer):
    @staticmethod
    def build_new_call_node(*, constant: str) -> ast.Call:
        return ast.Call(func=ast.Name(id="_", ctx=ast.Load()), args=[ast.Constant(constant)], keywords=[])

    def build_args_node(self, *, args: list[ast.Dict | ast.Constant | ast.expr]) -> list[ast.Dict | ast.Constant]:
        if args and isinstance(args[0], ast.Dict):
            dict_args: ast.Dict = args[0]
            param = dict_args.values[0]

            if hasattr(param, "func"):
                return args

            new_node = self.build_new_call_node(constant=param.value)
            args[0].values[0] = new_node
            return args
        if args and isinstance(args[0], ast.Constant):
            constant = args[0]
            new_node = self.build_new_call_node(constant=constant.value)
            args[0] = new_node
        return args

    def build_keywords_node_for_arg(self, *, keywords: list[ast.keyword], arg_name: str) -> list[ast.keyword]:
        if not keywords:
            return keywords
        message = next((keyword for keyword in keywords if keyword.arg == arg_name), None)
        if message and isinstance(message.value, ast.Constant):
            constant = message.value
            new_node = self.build_new_call_node(constant=constant.value)
            message.value = new_node
        return keywords

    def generate_tuple_gettext(self, *, value: ast.Tuple | stmt) -> ast.Tuple:
        last_constant = value.elts[-1]
        if isinstance(last_constant, ast.Constant):
            value.elts[-1] = self.build_new_call_node(constant=last_constant.value)
        return value

    def generate_class_boby_gettext(self, *, body: ast.Assign, instance: ast.ClassDef | stmt):
        if isinstance(body.value, ast.Constant):
            constant = body.value.value
            new_node = self.build_new_call_node(constant=constant)
            body.value = new_node

        if isinstance(body.value, ast.Tuple) and body.value.elts:
            if instance.name == "Meta":
                return None
            body.value = self.generate_tuple_gettext(value=body.value)

        return body

    def generate_class_gettext(self, *, instance: ast.ClassDef | stmt) -> ast.ClassDef:
        for body in instance.body:
            if not isinstance(body, ast.Assign) or body.targets[-1].id == "abstract":
                continue
            self.generate_class_boby_gettext(body=body, instance=instance)

        return instance

    def append_verbose_name(self, *, instance: ast.Assign | stmt) -> ast.Assign:
        constant = instance.targets[-1]
        new_node = self.build_new_call_node(constant=constant.id.title())
        instance.value.keywords.insert(0, ast.keyword(arg="verbose_name", value=new_node))

        return instance

    def generate_fk_gettext(self, *, instance: ast.Assign | stmt) -> ast.Assign:
        verbose = next((keyword for keyword in instance.value.keywords if keyword.arg == "verbose_name"), None)
        if verbose is None:
            self.append_verbose_name(instance=instance)
        else:
            constant = verbose.value
            if not isinstance(constant, ast.Constant):
                return instance
            new_node = self.build_new_call_node(constant=constant.value)
            verbose.value = new_node
        return instance

    def generate_assign_args_gettext(self, *, instance: ast.Assign | stmt) -> ast.Assign:
        if not instance.value.args and not instance.value.keywords:
            constant = instance.targets[-1]
            new_node = self.build_new_call_node(constant=constant.id.title())
            instance.value.args.append(new_node)
            return instance

        if instance.value.args:
            instance.value.args = self.build_args_node(args=instance.value.args)
        return instance

    def generate_keyword_gettext(self, *, keyword: ast.keyword) -> ast.keyword:
        if keyword.arg in ("verbose_name", "help_text"):
            constant = keyword.value
            if not isinstance(constant, ast.Constant):
                return keyword
            new_node = self.build_new_call_node(constant=constant.value)
            keyword.value = new_node
        return keyword

    def generate_assign_keywords_gettext(self, *, instance: ast.Assign | stmt) -> ast.Assign:
        if not instance.value.keywords:
            return instance

        for keyword in instance.value.keywords:
            self.generate_keyword_gettext(keyword=keyword)

        verbose = next((keyword for keyword in instance.value.keywords if keyword.arg == "verbose_name"), None)
        if not instance.value.args and verbose is None:
            self.append_verbose_name(instance=instance)

        return instance

    def generate_assign_gettext(self, *, instance: ast.Assign | stmt) -> ast.Assign:
        if instance.targets[-1].id == "objects":
            return instance

        with suppress(AttributeError):
            f_keys = ("ForeignKey", "ManyToManyField", "OneToOneField")
            func = instance.value.func
            if (hasattr(func, "attr") and func.attr in f_keys) or hasattr(func, "id"):
                self.generate_fk_gettext(instance=instance)
                return instance

        if isinstance(instance.value, ast.Tuple):
            instance.value = self.generate_tuple_gettext(value=instance.value)
            return instance

        try:
            self.generate_assign_args_gettext(instance=instance)
        except AttributeError:
            return instance

        self.generate_assign_keywords_gettext(instance=instance)

        return instance

    def generate_decorator_gettext(self, *, decorator: ast.Call | stmt, instance_name: str) -> ast.Call:
        for keyword in decorator.keywords:
            if keyword.arg == "description" and isinstance(keyword.value, ast.Constant):
                constant = keyword.value
                new_node = self.build_new_call_node(constant=constant.value.title())
                keyword.value = new_node

        description = next((keyword for keyword in decorator.keywords if keyword.arg == "description"), None)
        if description is None:
            decorator.keywords.append(
                ast.keyword(
                    arg="description",
                    value=self.build_new_call_node(constant=instance_name.replace("_", " ").title()),
                )
            )
        return decorator

    def generate_funcdef_decorator_gettext(self, *, instance: ast.FunctionDef | stmt) -> ast.FunctionDef:
        decorators = instance.decorator_list
        for decorator in decorators:
            if not isinstance(decorator, ast.Call):
                continue
            self.generate_decorator_gettext(decorator=decorator, instance_name=instance.name)

        return instance

    def generate_elt_gettext(self, *, elts: list[ast.Call]) -> list[ast.Call]:
        for elt in elts:
            if elt.args:
                elt.args = self.build_args_node(args=elt.args)
            elif elt.keywords:
                elt.keywords = self.build_keywords_node_for_arg(keywords=elt.keywords, arg_name="message")
        return elts

    def generate_raise_gettext(self, *, instance: ast.Raise | stmt) -> ast.Raise:
        exc = instance.exc

        if hasattr(exc, "elts") and exc.elts:
            self.generate_elt_gettext(elts=exc.elts)
        if hasattr(exc, "args") and exc.args:
            exc.args = self.build_args_node(args=exc.args)
            return instance
        if hasattr(exc, "keywords") and exc.keywords:
            exc.keywords = self.build_keywords_node_for_arg(keywords=exc.keywords, arg_name="message")
            return instance

        return instance

    def generate_funcdef_raising_gettext(self, *, instance: ast.FunctionDef | stmt) -> ast.FunctionDef:
        for body in instance.body:
            for item in body.body:
                if not isinstance(item, ast.Raise):
                    continue
                self.generate_raise_gettext(instance=item)

        return instance

    @staticmethod
    def insert_getetxt_import(node: ast.Module) -> ast.Module:
        aliases = [body.names[-1].name for body in node.body if isinstance(body, ast.ImportFrom)]
        if "gettext_lazy" in aliases:
            return node

        counted_imports = len([
            import_node for import_node in node.body if isinstance(import_node, (ast.ImportFrom, ast.Import))
        ])

        import_node = ast.ImportFrom(
            module="django.utils.translation",
            names=[ast.alias(name="gettext_lazy", asname="_")],
            level=0,
        )
        node.body.insert(counted_imports, import_node)
        return node

    def visit_ClassDef(self, node: ast.ClassDef) -> ast.ClassDef:  # noqa: N802
        match node:
            case ast.ClassDef(bases=[ast.Attribute(attr="ModelAdmin")]):
                for instance in node.body:
                    match instance:
                        case ast.FunctionDef(decorator_list=[ast.Call(keywords=[*_])]):
                            self.generate_funcdef_decorator_gettext(instance=instance)

            case ast.ClassDef(bases=[ast.Attribute(attr="TextChoices")]):
                return self.generate_class_gettext(instance=node)

            case ast.ClassDef(bases=[ast.Name(id=_), *_]):
                for instance in node.body:
                    match instance:
                        case ast.ClassDef():
                            self.generate_class_gettext(instance=instance)
                        case ast.Assign():
                            self.generate_assign_gettext(instance=instance)
                        case ast.FunctionDef(body=[ast.If()]):
                            self.generate_funcdef_raising_gettext(instance=instance)

            case ast.ClassDef(bases=[ast.Attribute(attr=a)]) if a == "Model":
                for instance in node.body:
                    match instance:
                        case ast.Assign():
                            self.generate_assign_gettext(instance=instance)

        return node



import re
from pathlib import Path

from deep_translator import GoogleTranslator
from deep_translator.exceptions import LanguageNotSupportedException
from django.conf import settings

from django_translate_gettext.exceptions import TranslatorError


class PoFileTranslator:
    def __init__(self, lang_code: str):
        self.lang_code = lang_code
        self.locale_paths = [Path(filepath) for filepath in settings.LOCALE_PATHS]
        try:
            self.translator = GoogleTranslator(source="auto", target=lang_code)
        except LanguageNotSupportedException as error:
            raise TranslatorError(f"Language code {lang_code} is not supported by the translator") from error

    def translate_block(self, block: str, msgid: list[str]) -> str:
        msgstr = re.findall(r'msgstr "(.*?)"', block)
        if msgstr and msgstr[0]:
            return block
        translated = self.translator.translate(msgid[0])
        return re.sub(r'msgstr "(.*?)*"', f'msgstr "{translated}"', block)

    def process_first_block(self, block: str) -> str:
        parts = block.split("\n")
        raw_msgid, raw_msgstr = parts[-2], parts[-1]
        msgid = re.findall(r'msgid "(.*?)"', raw_msgid)
        parts[-1] = self.translate_block(block=raw_msgstr, msgid=msgid)
        return "\n".join(parts)

    def translate_locale_path(self, *, locale_path: Path) -> None:
        result = []
        po_file = locale_path.joinpath(self.lang_code, "LC_MESSAGES", "django.po")
        if not po_file.exists():
            raise TranslatorError(f"The file for code {self.lang_code} does not exist.")

        file_content = po_file.read_text().split("\n\n")
        for block in file_content:
            if not block:
                continue

            msgid = re.findall(r'msgid "(.*?)"', block)
            if len(msgid) != 1:
                result.append(self.process_first_block(block=block))
                continue

            result.append(self.translate_block(block=block, msgid=msgid))

        po_file.write_text("\n\n".join(result))

    def translate_codes(self) -> None:
        for locale_path in self.locale_paths:
            self.translate_locale_path(locale_path=locale_path)
